/*
-----------------------------------------------------------------------------
This source file is part of OGRE
    (Object-oriented Graphics Rendering Engine)
For the latest info, see http://ogre.sourceforge.net/

Copyright  2000-2002 The OGRE Team
Also see acknowledgements in Readme.html

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General  License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General  License for more details.

You should have received a copy of the GNU Lesser General  License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
-----------------------------------------------------------------------------
*/


#import <Cocoa/Cocoa.h>
#import "OgreOSXConfigWindowController.h"
#include "OgreOSXConfig.h"



using namespace Ogre;

bool OSXConfig::display(void)
{
    OgreConfigWindowController *config = [[OgreConfigWindowController alloc] 
                                            initWithWindowNibName: [NSString stringWithCString: "config"]];
	[config setup];

	// Just pick first renderer
    RenderSystemList* renderers = Root::getSingleton().getAvailableRenderers();
    RenderSystem* renderer = renderers->front();

    ConfigOptionMap options = renderer->getConfigOptions();

    // Setup Render Options
	StringVector::iterator opt_it;
	ConfigOption videoModes = options.find("Video Mode")->second;
	for(opt_it = videoModes.possibleValues.begin(); opt_it != videoModes.possibleValues.end(); opt_it++)
	{
		NSLog(@"ADD %@", [NSString stringWithCString: (*opt_it).c_str()]);
		[config addRenderOption: [NSString stringWithCString: (*opt_it).c_str()]];
	}
    // Setup FSAA
	ConfigOption fsaaModes = options.find("FSAA")->second;
	for(opt_it = fsaaModes.possibleValues.begin(); opt_it != fsaaModes.possibleValues.end(); opt_it++)
	{
		[config addFSAAOption: [NSString stringWithCString: (*opt_it).c_str()]];
	}
	
	
	[config performSelectorOnMainThread:@selector(runConfigure) withObject:NULL waitUntilDone:YES];
	
	if(![config done])
		return false;


	// Get Options
	renderer->setConfigOption("Video Mode", videoModes.possibleValues[[config renderOption]]);
	renderer->setConfigOption("FSAA", fsaaModes.possibleValues[[config fsaaOption]]);
	if([config fullScreen])
		renderer->setConfigOption("Full Screen", "Yes");
	
    // All done
    Root::getSingleton().setRenderSystem(renderer);
    Root::getSingleton().saveConfig();    
    [config release];
    return true;
}


@implementation OgreConfigWindowController : NSWindowController

- (void)setup
{
	[self window];
    [resolutionPopUp removeAllItems];
    [FSAAPopUp removeAllItems];
}

- (void)addRenderOption:(NSString *)title
{
    [resolutionPopUp addItemWithTitle:title];
}

- (void)addFSAAOption:(NSString *)title
{
    [FSAAPopUp addItemWithTitle:title];
}

- (int)renderOption
{
    return renderChoice;
}

- (int)fsaaOption
{
    return fsaaChoice;
}

- (BOOL)fullScreen
{
    return [fullScreenCheckBox state] == NSOnState;
}

- (BOOL)done
{
    return done;
}

- (void)okClicked:(id)sender
{
    [[NSApplication sharedApplication] stopModal];
	done = YES;
}

- (void)cancelClicked:(id)sender
{
    [[NSApplication sharedApplication] stopModal];
}

- (void)runConfigure
{
    [self showWindow: self];

	done = NO;
    [[NSApplication sharedApplication] runModalForWindow: [self window]];

	renderChoice = [resolutionPopUp indexOfSelectedItem];
	fsaaChoice = [FSAAPopUp indexOfSelectedItem];
    [[self window] close];
}

@end