/*
-----------------------------------------------------------------------------
This source file is part of OGRE
    (Object-oriented Graphics Rendering Engine)
For the latest info, see http://www.ogre3d.org/

Copyright (c) 2000-2005 The OGRE Team
Also see acknowledgements in Readme.html

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
-----------------------------------------------------------------------------
*/

#include "OgreOSXInput.h"
#include "OgreOSXWindow.h"
#include "OgreLogManager.h"
#include "OgreMouseEvent.h"
#include "OgreCursor.h"
#include "OgreEventQueue.h"
#import <Cocoa/Cocoa.h>

	
namespace Ogre {

    OSXInput::OSXInput() : InputReader()
    {
		mScale = 0.002;

		useBufferedInput(new EventQueue(), true, true);

        _key_map.insert(InputKeyMap::value_type(0x35, KC_ESCAPE));
        _key_map.insert(InputKeyMap::value_type(0x12, KC_1));
        _key_map.insert(InputKeyMap::value_type(0x13, KC_2));
        _key_map.insert(InputKeyMap::value_type(0x14, KC_3));
        _key_map.insert(InputKeyMap::value_type(0x15, KC_4));
        _key_map.insert(InputKeyMap::value_type(0x17, KC_5));
        _key_map.insert(InputKeyMap::value_type(0x16, KC_6));
        _key_map.insert(InputKeyMap::value_type(0x1a, KC_7));
        _key_map.insert(InputKeyMap::value_type(0x1c, KC_8));
        _key_map.insert(InputKeyMap::value_type(0x19, KC_9));
        _key_map.insert(InputKeyMap::value_type(0x1d, KC_0));
        _key_map.insert(InputKeyMap::value_type(0x1b, KC_MINUS));
        _key_map.insert(InputKeyMap::value_type(0x18, KC_EQUALS));
        _key_map.insert(InputKeyMap::value_type(0x33, KC_BACK));
        _key_map.insert(InputKeyMap::value_type(0x30, KC_TAB));
        _key_map.insert(InputKeyMap::value_type(0x0c, KC_Q));
        _key_map.insert(InputKeyMap::value_type(0x0d, KC_W));
        _key_map.insert(InputKeyMap::value_type(0x0e, KC_E));
        _key_map.insert(InputKeyMap::value_type(0x0f, KC_R));
        _key_map.insert(InputKeyMap::value_type(0x11, KC_T));
        _key_map.insert(InputKeyMap::value_type(0x10, KC_Y));
        _key_map.insert(InputKeyMap::value_type(0x20, KC_U));
        _key_map.insert(InputKeyMap::value_type(0x22, KC_I));
        _key_map.insert(InputKeyMap::value_type(0x1f, KC_O));
        _key_map.insert(InputKeyMap::value_type(0x23, KC_P));
        _key_map.insert(InputKeyMap::value_type(0x21, KC_LBRACKET));
        _key_map.insert(InputKeyMap::value_type(0x1e, KC_RBRACKET));
        _key_map.insert(InputKeyMap::value_type(0x24, KC_RETURN));
        _key_map.insert(InputKeyMap::value_type(0x3b, KC_LCONTROL));
        _key_map.insert(InputKeyMap::value_type(0x00, KC_A));
        _key_map.insert(InputKeyMap::value_type(0x01, KC_S));
        _key_map.insert(InputKeyMap::value_type(0x02, KC_D));
        _key_map.insert(InputKeyMap::value_type(0x03, KC_F));
        _key_map.insert(InputKeyMap::value_type(0x05, KC_G));
        _key_map.insert(InputKeyMap::value_type(0x04, KC_H));
        _key_map.insert(InputKeyMap::value_type(0x26, KC_J));
        _key_map.insert(InputKeyMap::value_type(0x28, KC_K));
        _key_map.insert(InputKeyMap::value_type(0x25, KC_L));
        _key_map.insert(InputKeyMap::value_type(0x29, KC_SEMICOLON));
//        _key_map.insert(InputKeyMap::value_type(0x27, KC_COLON));
        _key_map.insert(InputKeyMap::value_type(0x27, KC_APOSTROPHE));
        _key_map.insert(InputKeyMap::value_type(0x32, KC_GRAVE));
        _key_map.insert(InputKeyMap::value_type(0x38, KC_LSHIFT));
        _key_map.insert(InputKeyMap::value_type(0x2a, KC_BACKSLASH));
        _key_map.insert(InputKeyMap::value_type(0x06, KC_Z));
        _key_map.insert(InputKeyMap::value_type(0x07, KC_X));
        _key_map.insert(InputKeyMap::value_type(0x08, KC_C));
        _key_map.insert(InputKeyMap::value_type(0x09, KC_V));
        _key_map.insert(InputKeyMap::value_type(0x0b, KC_B));
        _key_map.insert(InputKeyMap::value_type(0x2d, KC_N));
        _key_map.insert(InputKeyMap::value_type(0x2e, KC_M));
        _key_map.insert(InputKeyMap::value_type(0x2b, KC_COMMA));
        _key_map.insert(InputKeyMap::value_type(0x2f, KC_PERIOD));
        _key_map.insert(InputKeyMap::value_type(0x3c, KC_RSHIFT));
//        _key_map.insert(InputKeyMap::value_type('*', KC_MULTIPLY));
        _key_map.insert(InputKeyMap::value_type(0x3b, KC_LMENU));
        _key_map.insert(InputKeyMap::value_type(0x31, KC_SPACE));
        _key_map.insert(InputKeyMap::value_type(0x39, KC_CAPITAL));
        _key_map.insert(InputKeyMap::value_type(0x7a, KC_F1));
        _key_map.insert(InputKeyMap::value_type(0x78, KC_F2));
        _key_map.insert(InputKeyMap::value_type(0x63, KC_F3));
        _key_map.insert(InputKeyMap::value_type(0x76, KC_F4));
        _key_map.insert(InputKeyMap::value_type(0x60, KC_F5));
        _key_map.insert(InputKeyMap::value_type(0x61, KC_F6));
        _key_map.insert(InputKeyMap::value_type(0x62, KC_F7));
        _key_map.insert(InputKeyMap::value_type(0x64, KC_F8));
        _key_map.insert(InputKeyMap::value_type(0x65, KC_F9));
        _key_map.insert(InputKeyMap::value_type(0x6d, KC_F10));
        _key_map.insert(InputKeyMap::value_type(0x47, KC_NUMLOCK));
//        _key_map.insert(InputKeyMap::value_type(NSScrollLockFunctionKey, KC_SCROLL));
        _key_map.insert(InputKeyMap::value_type(0x59, KC_NUMPAD7));
        _key_map.insert(InputKeyMap::value_type(0x5b, KC_NUMPAD8));
        _key_map.insert(InputKeyMap::value_type(0x5c, KC_NUMPAD9));
        _key_map.insert(InputKeyMap::value_type(0x4e, KC_SUBTRACT));
        _key_map.insert(InputKeyMap::value_type(0x56, KC_NUMPAD4));
        _key_map.insert(InputKeyMap::value_type(0x57, KC_NUMPAD5));
        _key_map.insert(InputKeyMap::value_type(0x58, KC_NUMPAD6));
        _key_map.insert(InputKeyMap::value_type(0x45, KC_ADD));
        _key_map.insert(InputKeyMap::value_type(0x53, KC_NUMPAD1));
        _key_map.insert(InputKeyMap::value_type(0x54, KC_NUMPAD2));
        _key_map.insert(InputKeyMap::value_type(0x55, KC_NUMPAD3));
        _key_map.insert(InputKeyMap::value_type(0x52, KC_NUMPAD0));
        _key_map.insert(InputKeyMap::value_type(0x41, KC_DECIMAL));
        _key_map.insert(InputKeyMap::value_type(0x67, KC_F11));
        _key_map.insert(InputKeyMap::value_type(0x6f, KC_F12));
        _key_map.insert(InputKeyMap::value_type(0x69, KC_F13));
//        _key_map.insert(InputKeyMap::value_type(NSF14FunctionKey, KC_F14));
//        _key_map.insert(InputKeyMap::value_type(NSF15FunctionKey, KC_F15));
        _key_map.insert(InputKeyMap::value_type(0x51, KC_NUMPADEQUALS));
        _key_map.insert(InputKeyMap::value_type(0x4b, KC_DIVIDE));
//        _key_map.insert(InputKeyMap::value_type(NSSysReqFunctionKey, KC_SYSRQ));
        _key_map.insert(InputKeyMap::value_type(0x3e, KC_RMENU));
        _key_map.insert(InputKeyMap::value_type(0x73, KC_HOME));
        _key_map.insert(InputKeyMap::value_type(0x77, KC_END));
        _key_map.insert(InputKeyMap::value_type(0x74, KC_PGUP));
        _key_map.insert(InputKeyMap::value_type(0x79, KC_PGDOWN));
        _key_map.insert(InputKeyMap::value_type(0x7e, KC_UP));
        _key_map.insert(InputKeyMap::value_type(0x7b, KC_LEFT));
        _key_map.insert(InputKeyMap::value_type(0x7c, KC_RIGHT));
        _key_map.insert(InputKeyMap::value_type(0x7d, KC_DOWN));
//        _key_map.insert(InputKeyMap::value_type(NSInsertFunctionKey, KC_INSERT));
        _key_map.insert(InputKeyMap::value_type(0x75, KC_DELETE));
        _key_map.insert(InputKeyMap::value_type(0x37, KC_LWIN));
        _key_map.insert(InputKeyMap::value_type(0x36, KC_RWIN));
    }

    OSXInput::~OSXInput()
    {
/*		if ( mUseMouse )
		{
			SDL_WM_GrabInput(SDL_GRAB_OFF);
			SDL_ShowCursor(SDL_ENABLE);
		}*/
    }

    void OSXInput::initialise(RenderWindow* pWindow, bool useKeyboard, bool useMouse, bool useGameController)
    {
		mUseMouse = useMouse;
		mView = ((OSXWindow*)pWindow)->ogreView();
    }

	void OSXInput::_releaseMouse()
	{
/*		SDL_WM_GrabInput(SDL_GRAB_OFF);
		SDL_ShowCursor(SDL_ENABLE);
		mMouseGrabbed = false;
		mMouseLeft = false;
		mGrabMouse = true;*/
	}

	void OSXInput::_grabMouse()
	{
/*		SDL_ShowCursor(SDL_DISABLE);
		SDL_WM_GrabInput(SDL_GRAB_ON);
		mMouseGrabbed = true;
		mGrabMouse = false;*/
	}

    void OSXInput::capture()
    {
		// Keyboard input
		if (mUseBufferedKeys)
        {
            processBufferedKeyboard();
        }

        if (mUseMouse && mUseBufferedMouse)
        {
            processBufferedMouse();
        }
/*		else
		{
            mMouseKeys = 0;
            mMouseRelativeX = 0; mMouseRelativeY = 0; mMouseRelativeZ = 0;

            // Get mouse info
            if( 1 ) // SDL_GetAppState() & SDL_APPMOUSEFOCUS )
            {
               mMouseKeys = SDL_GetMouseState( &mMouseX, &mMouseY );
                SDL_GetRelativeMouseState( &mMouseRelativeX, &mMouseRelativeY );


                // get mouse wheel movement
                // fetch all mouse wheel related events
                int count = SDL_PeepEvents( event, 16 , SDL_GETEVENT
						,(SDL_EVENTMASK(SDL_MOUSEBUTTONDOWN)
						| SDL_EVENTMASK(SDL_MOUSEBUTTONUP)
						| SDL_EVENTMASK(SDL_MOUSEMOTION)
					   ));
                if ( count )
                {
                    for (int i = 0; i < count; i++)
                    {
						switch (event[i].type)
						{
						case SDL_MOUSEBUTTONDOWN:
						case SDL_MOUSEBUTTONUP:
							// grab the mouse if the user presses a mouse button
							if ( !mMouseGrabbed && mGrabMouse && ( mGrabMode == GRAB_MOUSE_CLICK ) )
							{
								_grabMouse();
							}
							switch(event[i].button.button)
							{
							case SDL_BUTTON_WHEELUP:
								mMouseRelativeZ += mWheelStep;
								break;
							case SDL_BUTTON_WHEELDOWN:
								mMouseRelativeZ -= mWheelStep;
								break;
							}
							break;
						}
                    }
                }
				
				NSPoint loc = [NSEvent mouseLocation];
				mMouseRelativeX = loc.x;
				mMouseRelativeY = loc.y;
            }

            mMouseState.Xabs = mMouseX;
            mMouseState.Yabs = mMouseY;
            mMouseState.Zabs = 0;

            mMouseState.Xrel = mMouseRelativeX;
            mMouseState.Yrel = mMouseRelativeY;
            mMouseState.Zrel = mMouseRelativeZ;

            mMouseState.Buttons =  (mMouseKeys & SDL_BUTTON_LMASK) ? 1 : 0; // left
            mMouseState.Buttons |= (mMouseKeys & SDL_BUTTON_RMASK) ? 2 : 0; // right
            mMouseState.Buttons |= (mMouseKeys & SDL_BUTTON_MMASK) ? 4 : 0; // middle

        }*/

		// XXX Fix me up
		// Game controller state
    }

    bool OSXInput::isKeyDownImmediate(KeyCode kc) const
    {
		return (mBufferedKeysDown.find(kc) != 0);
    }

    long OSXInput::getMouseRelX() const
    {
        return mMouseState.Xrel;
    }

    long OSXInput::getMouseRelY() const
    {
        return mMouseState.Yrel;
    }

    long OSXInput::getMouseRelZ() const
    {
        return mMouseState.Zrel;
    }

    long OSXInput::getMouseAbsX() const
    {
        return mMouseState.Xabs;
    }

    long OSXInput::getMouseAbsY() const
    {
        return mMouseState.Yabs;
    }

    long OSXInput::getMouseAbsZ() const
    {
        return 0;
    }

    bool OSXInput::getMouseButton( uchar button ) const
    {
        return mMouseState.isButtonDown( button );
    }

    void OSXInput::processBufferedKeyboard()
    {
		//slow but it'll do for now
		//[[mView getLock] lock];
		while([[mView keyEvents] count] > 0)
		{
			NSEvent *theEvent = [[mView keyEvents] objectAtIndex:0];
			
			if([theEvent type] == NSFlagsChanged)
			{
				bool down;
				if(mBufferedKeysDown.find(_key_map[[theEvent keyCode]]) == 0)
					down = true;
				else
					down = false;
				keyChanged(_key_map[[theEvent keyCode]], down);
			}
			else
			{
				if([theEvent type] == NSKeyDown)
				{
					keyChanged(_key_map[[theEvent keyCode]], true);
				}
				else if([theEvent type] == NSKeyUp)
				{
					keyChanged(_key_map[[theEvent keyCode]], false);
				}
			}
			[[mView keyEvents] removeObjectAtIndex:0];
		}
		//[[mView getLock] unlock];
    }

    void OSXInput::processBufferedMouse()
    {
		// Mouse Events
		//slow but it'll do for now
		//[[mView getLock] lock];
		while([[mView mouseEvents] count] > 0)
		{
			NSEvent *theEvent = [[mView mouseEvents] objectAtIndex:0];
			
            int button_mask = -1;
            bool button_down = false;
			
			switch([theEvent type])
			{
			case NSLeftMouseDown:
			case NSRightMouseDown:
			case NSOtherMouseDown:
				button_down = true;
			case NSLeftMouseUp:
			case NSRightMouseUp:
			case NSOtherMouseUp:

                switch([theEvent buttonNumber])
                {
                case 0:
                    button_mask = InputEvent::BUTTON0_MASK;
                    break;
                case 1:
                    button_mask = InputEvent::BUTTON1_MASK;
                    break;
                case 2:
                    button_mask = InputEvent::BUTTON2_MASK;
                    break;
                };
                triggerMouseButton(button_mask, button_down);
                break;
			case NSScrollWheel:
				//mMouseState.Xrel += [theEvent deltaX];
				//mMouseState.Yrel += [theEvent deltaY];
				mMouseState.Zrel += [theEvent deltaY];
				break;
			default:
				break;
			};
			
			
			[[mView mouseEvents] removeObjectAtIndex:0];
		}
		//[[mView getLock] unlock];
		
		
		// Location
		NSPoint loc = [NSEvent mouseLocation];
		mMouseState.Xrel = (loc.x - mMouseState.Xabs)* mScale;
		mMouseState.Yrel = (loc.y - mMouseState.Yabs)* mScale;
		mMouseState.Xabs = loc.x;
		mMouseState.Yabs = loc.y;
		mMouseState.Zabs = 0;
		/*
		mCursor->addToX(loc.x * mScale);
		mCursor->addToY(loc.y * mScale);
		mouseMoved();*/
			
        // XXX Arbitrarily picked 16 
       /* SDL_Event events[16];

        int count = SDL_PeepEvents(events, 16, SDL_GETEVENT,
                (SDL_MOUSEMOTIONMASK | SDL_MOUSEBUTTONDOWNMASK |
                 SDL_MOUSEBUTTONUPMASK | SDL_ACTIVEEVENTMASK));
        if (!count)
        {
            return;
        }

        bool Xset, Yset, Zset;
        Xset = Yset = Zset = false;
        for (int i = 0; i < count; i++)
        {
            int button_mask = -1;
            bool button_down = false;
            switch (events[i].type)
            {
            case SDL_ACTIVEEVENT:
                _visible = events[i].active.gain ? true : false;
                break;
            case SDL_MOUSEMOTION:
                if (events[i].motion.xrel)
                {
                    if (Xset)
                    {
                        mouseMoved();
                    }
                    mCursor->addToX(events[i].motion.xrel * mScale);
                    Xset = true;
                }
                if (events[i].motion.yrel)
                {
                    if (Yset)
                    {
                        mouseMoved();
                    }
                    mCursor->addToY(events[i].motion.yrel * mScale);
                    Yset = true;
                }
                break;
            case SDL_MOUSEBUTTONDOWN:
                button_down = true;
            case SDL_MOUSEBUTTONUP:
				// grab the mouse if the user presses a mouse button
				if ( !mMouseGrabbed && mGrabMouse && ( mGrabMode == GRAB_MOUSE_CLICK ) )
				{
					_grabMouse();
				}
                switch(events[i].button.button)
                {
                case SDL_BUTTON_LEFT:
                    button_mask = InputEvent::BUTTON0_MASK;
                    break;
                case SDL_BUTTON_RIGHT:
                    button_mask = InputEvent::BUTTON1_MASK;
                    break;
                case SDL_BUTTON_MIDDLE:
                    button_mask = InputEvent::BUTTON2_MASK;
                    break;
                };
                triggerMouseButton(button_mask, button_down);
                break;
            };
            
            if (Xset && Yset)
            {
                mouseMoved();
                Xset = Yset = false;
            }
        }

        if (Xset || Yset)
        {
            mouseMoved();
        }*/
                    
    }

    void OSXInput::getMouseState( MouseState& state ) const
    {
        memcpy( &state, &mMouseState, sizeof( MouseState ) );
    }

}
