#include "OgreException.h"
#include "OgreLogManager.h"
#include "OgreStringConverter.h"

#include "OgreOSXGLSupport.h"
#include "OgreOSXWindow.h"
#include <OpenGL/OpenGL.h>

using namespace Ogre;

OSXGLSupport::OSXGLSupport()
{

    //SDL_Init(SDL_INIT_VIDEO);
}

OSXGLSupport::~OSXGLSupport()
{
}

void OSXGLSupport::addConfig(void)
{
   /* mVideoModes = SDL_ListModes(NULL, SDL_FULLSCREEN | SDL_OPENGL);
    
    if (mVideoModes == (SDL_Rect **)0)
    {
        OGRE_EXCEPT(999, "Unable to load video modes",
                "SDLRenderSystem::initConfigOptions");
    }

    ConfigOption optFullScreen;
    ConfigOption optVideoMode;
    ConfigOption optFSAA;

    // FS setting possiblities
    optFullScreen.name = "Full Screen";
    optFullScreen.possibleValues.push_back("Yes");
    optFullScreen.possibleValues.push_back("No");
    optFullScreen.currentValue = "Yes";
    optFullScreen.immutable = false;
    */
    //FSAA possibilities
	ConfigOption optFullScreen;
    ConfigOption optVideoMode;
    ConfigOption optFSAA;
	
	// Get renderers
	CGLRendererInfoObj rend;
	long nrend;
	CGLQueryRendererInfo(CGDisplayIDToOpenGLDisplayMask(kCGDirectMainDisplay), &rend, &nrend);
	
	long maxSamples;
	CGLDescribeRenderer(rend, 0, kCGLRPMaxSamples, &maxSamples);
	
	
	optFSAA.name = "FSAA";
	optFSAA.possibleValues.push_back("0");
	
	switch(maxSamples)
	{
		case 6:
			optFSAA.possibleValues.push_back("2");
			optFSAA.possibleValues.push_back("4");
			optFSAA.possibleValues.push_back("6");
			break;
		case 4:
			optFSAA.possibleValues.push_back("2");
			optFSAA.possibleValues.push_back("4");
			break;
		case 2:
			optFSAA.possibleValues.push_back("2");
			break;
		default: break;
	}
	
    optFSAA.currentValue = "0";
    optFSAA.immutable = false;
    
    mOptions[optFSAA.name] = optFSAA;

     // FS setting possiblities
    optFullScreen.name = "Full Screen";
    optFullScreen.possibleValues.push_back("Yes");
    optFullScreen.possibleValues.push_back("No");
    optFullScreen.currentValue = "No";
    optFullScreen.immutable = false;
 
    // Video mode possiblities
    optVideoMode.name = "Video Mode";
    optVideoMode.immutable = false;

	CFArrayRef display_modes = CGDisplayAvailableModes(CGMainDisplayID());
	CFIndex num_modes = CFArrayGetCount(display_modes);
	NSMutableArray *resolutions = [NSMutableArray array];

	for (int i = 0; i < num_modes; i++)
	{	
		NSDictionary *modeInfo = (NSDictionary*)CFArrayGetValueAtIndex(display_modes, i);
		
		if([modeInfo objectForKey:(NSString*)kCGDisplayModeIsSafeForHardware] != NULL)
		{
			NSString *res = [NSString stringWithFormat:@"%@ x %@", [modeInfo objectForKey:(NSString*)kCGDisplayWidth],
				[modeInfo objectForKey:(NSString*)kCGDisplayHeight]];
			
			if(![resolutions containsObject:res])
				[resolutions addObject:res];
			NSLog(@"HAS RES: %@ %@ %@",res, [modeInfo objectForKey:(NSString*)kCGDisplayRefreshRate], [modeInfo objectForKey:(NSString*)kCGDisplayBitsPerPixel]);
		}
	}
	
	//Resolutions
	[resolutions sortUsingSelector:@selector(compare:)];
	int num_res = [resolutions count];
	for (int i = 0; i < num_res; i++)
	{	
		optVideoMode.possibleValues.push_back([[resolutions objectAtIndex:i] cString]);
	}

    mOptions[optFullScreen.name] = optFullScreen;
    mOptions[optVideoMode.name] = optVideoMode;
}

String OSXGLSupport::validateConfig(void)
{
    return String("");
}

RenderWindow* OSXGLSupport::createWindow(bool autoCreateWindow, GLRenderSystem* renderSystem, const String& windowTitle)
{
	if (autoCreateWindow)
    {
		ConfigOptionMap::iterator opt = mOptions.find("Full Screen");
        bool fullscreen = (opt->second.currentValue == "Yes");
	
        String val = mOptions.find("Video Mode")->second.currentValue;
        String::size_type pos = val.find('x');
        unsigned int w = StringConverter::parseUnsignedInt(val.substr(0, pos));
        unsigned int h = StringConverter::parseUnsignedInt(val.substr(pos + 1));
        return renderSystem->createRenderWindow(windowTitle, w, h, fullscreen, 0);
    }
    else
    {
        // XXX What is the else?
		return NULL;
    }

}


RenderWindow* OSXGLSupport::newWindow(const String &name, unsigned int width, unsigned int height, 
	bool fullScreen, const NameValuePairList *miscParams)
{
	OSXWindow* window = new OSXWindow();
	window->create(name, width, height, fullScreen, miscParams);
	return window;
}

void OSXGLSupport::start()
{
    LogManager::getSingleton().logMessage(
        "********************************\n"
        "*** Starting GLOSX Subsystem ***\n"
        "********************************");
}

void OSXGLSupport::stop()
{
    LogManager::getSingleton().logMessage(
        "********************************\n"
        "*** Stopping GLOSX Subsystem ***\n"
        "********************************");
}

#import <mach-o/dyld.h>
#import <stdlib.h>
#import <string.h>

void *OSXGLSupport::getProcAddress(const char *name)
{
    NSSymbol symbol;
    char *symbolName;
    // Prepend a '_' for the Unix C symbol mangling convention
    symbolName = (char*)malloc (strlen (name) + 2);
    strcpy(symbolName + 1, name);
    symbolName[0] = '_';
    symbol = NULL;
    if (NSIsSymbolNameDefined (symbolName))
        symbol = NSLookupAndBindSymbol (symbolName);
    free (symbolName);
    return symbol ? NSAddressOfSymbol (symbol) : NULL;
}

void *OSXGLSupport::getProcAddress(const String& procname)
{
   return getProcAddress(procname.c_str());
}